---
title: "OT"
author: "Jean Feydy"
institute: ""
topic: ""
theme: "metropolis"
colortheme:
fonttheme:
fontsize: 11pt
linkcolor: black
urlcolor: blue
linkstyle: bold
aspectratio: 169
titlegraphic: 
logo:
date:
section-titles: false
toc: false
---


## Who am I?

Background in **mathematics** and **data sciences**:

```{=latex}
\vspace*{-.25cm}
\begin{description}
\setlength\itemsep{-.5em}
\item[2012--2016] ENS Paris, mathematics.
\item[2014--2015] M2 mathematics, vision, learning at ENS Cachan.
\item[2016--2019] PhD thesis in \textcolor{red!80!black}{\textbf{medical imaging}} with Alain Trouvé at ENS Cachan.
\item[2019--2021] \textcolor{red!80!black}{\textbf{Geometric deep learning}} with Michael Bronstein at Imperial College.
\item[2021+] \textcolor{red!80!black}{\textbf{Medical data analysis}} in the HeKA INRIA team (Paris).
\end{description}
\vspace*{.4cm}
```


## HeKA : a translational research team for public health

`\begin{center}`{=latex}

![](images/HeKA.pdf){height=7cm}

`\end{center}`{=latex}



## My main motivation

Develop **robust and efficient** software that $\BR{stimulates}$ **other researchers**:

1. Speed up $\BR{geometric}$ **machine learning** on GPUs: \
   $\Longrightarrow~~$ **pyKeOps** library for distance and kernel matrices, 600k+ downloads.

2. Scale up $\BR{pharmacovigilance}$ to the full French population: \
   $\Longrightarrow~~$ **survivalGPU**, a fast re-implementation of the R survival package.

3. Ease access to modern statistical $\BR{shape analysis}$: \
   $\Longrightarrow~~$ **GeomLoss**, truly scalable optimal transport in Python. \
   $\Longrightarrow~~$ **scikit-shapes**, alpha release now available.


## Today’s talk -- assuming that you would enjoy some nice simulations

1. A quick heads up on **fast geometric methods**. \

2. Efficient discrete optimal transport **solvers**. \

3. New applications for systems of **incompressible particles**. \


# How to code a N-body simulation?

## Scientific computing libraries represent most objects as tensors

::: columns

:::: {.column width=70%}

`\vspace*{.25cm}`{=latex}
`\contexte{Context.}`{=latex}
Constrained **memory accesses** on the GPU:
`\vspace*{.3cm}`{=latex}

- **Long access times** to the registers \
  penalize the use of large **dense** arrays.

- Hard-wired **contiguous** memory accesses \
  penalize the use of **sparse** matrices.


\
`\defi{Challenge.}`{=latex}
In order to reach optimal run times:
`\vspace*{.3cm}`{=latex}

- **Restrict** ourselves to operations that are supported \
  by the constructor: convolutions, FFT, etc.

- Develop new routines from scratch in C++/CUDA \
  (FAISS, KPConv...): **several months of work**.


::::

:::: {.column width=25%}

`\begin{center}`{=latex}


```{=latex}
\begin{tikzpicture}[scale=.9]
	\draw[draw=black, fill=blue!20] (0,0) rectangle ++(3,3);
	\draw[step=.15, color=gray] (0,0) grid ++(3,3);
	\draw[draw=black, thick] (0,0) rectangle ++(3,3);
	\node[inner sep=3pt, anchor = mid, fill=blue!5, rounded corners=2] at (1.5,1.5) {{\normalsize M[\,i\,,\,j\,]}};
\end{tikzpicture}
```

**Dense array**

```{=latex}
\vspace*{.5cm}

\begin{tikzpicture}[scale=.9]
	\draw[draw=black, thick] (0,0) rectangle ++(3,3);
	\node[inner sep=0pt, anchor = mid] at (1.5,1.5) {{\normalsize ($\text{i}_\text{n}$, $\text{j}_\text{n}$, $\text{M}_\text{n}$)}};
	\draw[fill=blue!20] (4*.15,1*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (15*.15,1*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (2*.15,2*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (9*.15,3*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (16*.15,3*.15) rectangle ++(.15,.15);
	%
	%\draw[fill=blue!20] (3*.15,4*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (6*.15,4*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (4*.15,6*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (14*.15,7*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (2*.15,7*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (18*.15,8*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (1*.15,9*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (16*.15,10*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (6*.15,13*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (16*.15,13*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (4*.15,14*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (12*.15,15*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (2*.15,16*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (10*.15,17*.15) rectangle ++(.15,.15);
\end{tikzpicture}
```

**Sparse matrix**

`\end{center}`{=latex}

::::

:::

  
## The KeOps library: efficient support for symbolic matrices


::: columns

:::: {.column width=70%}

`\solution{Solution.}`{=latex} 
KeOps -- `\textcolor{blue}{\texttt{www.kernel-operations.io}}`{=latex}: `\vspace*{.2cm}`{=latex}

- For PyTorch, NumPy, Matlab and R, on **CPU and GPU**.
- **Automatic differentiation**.
- Just-in-time **compilation** of **optimized** C++ schemes, \
  triggered for every new **reduction**: sum, min, etc.

\
If the formula "F" is simple ($\leqslant$ 100 arithmetic operations): \
$\qquad$ "100k $\times$ 100k" computation  $\rightarrow$  10ms -- 100ms, \
$\qquad$     "1M $\times$ 1M"    computation  $\rightarrow\phantom{\text{0m}}$ 1s -- 10s.

`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}

Hardware ceiling of $\text{10}^{\text{12}}$ operations/s. \
**$\times \textbf{10}$ to $\times \textbf{100}$ speed-up** vs standard GPU implementations \
for a wide range of problems.

`\end{center}`{=latex}

::::

:::: {.column width=25%}

`\begin{center}`{=latex}



```{=latex}
\begin{tikzpicture}[scale=.9]
	% Left border:
	\draw[draw=black, fill=red!20] (-.45,0) rectangle ++(.3,3);
	\draw[draw=black, fill=red!80] (-.45,9*.15) rectangle ++(.3,.15);
	\draw[step=.15, color=gray] (-.45,0) grid ++(.3,3);
	\draw[draw=black] (-.45,0) rectangle ++(.3,3);
	%
	% Top border:
	\draw[draw=black, fill=green!20] (0,3.15) rectangle ++(3,.3);
	\draw[draw=black, fill=green!80] (6*.15,3.15) rectangle ++(.15,.3);
	\draw[step=.15, color=gray] (0,3.15) grid ++(3,.3);
	\draw[draw=green!50!black] (0,3.15) rectangle ++(3,.3);
	%
	\draw[draw=black,thick] (0,0) rectangle ++(3,3);
	%
	\draw[->, thick, color=red!80!black] (.5*.15, 9.5*.15) -- (5*.15, 9.5*.15);
	\draw[->, thick, color=green!50!black] (6.5*.15, 19.5*.15) -- (6.5*.15, 11*.15);
	%
	\draw[fill=blue!40] (6*.15,9*.15) rectangle ++(.15,.15);
	\node[inner sep=0pt, anchor = mid, color=blue!50!black] at (1.95,1.4) {\normalsize{F(\,\textcolor{red!80!black}{$\text{x}_\text{i}$}\,,\,\textcolor{green!50!black}{$\text{y}_\text{j}$}\,)}};
\end{tikzpicture}
```

**Symbolic matrix** \
Formula + data
 

`\end{center}`{=latex}
`\vspace*{.5cm}`{=latex}


- Distances `d(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Kernel `k(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Numerous transforms.


::::

:::


```{=latex}

\end{frame}

\begin{frame}[fragile]{A first example: efficient nearest neighbor search in dimension 50}

Create large point clouds using \textbf{standard PyTorch syntax}:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
import torch
N, M, D = 10**6, 10**6, 50
x = torch.rand(N, 1, D).cuda()  # (1M,  1, 50) array
y = torch.rand(1, M, D).cuda()  # ( 1, 1M, 50) array
\end{minted}

%\pause
\vspace*{-.6cm}
Turn \textbf{dense} arrays into \textbf{symbolic} matrices:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||,highlightlines={1-2}]{python}
from pykeops.torch import LazyTensor
x_i, y_j = LazyTensor(x), LazyTensor(y)
\end{minted}

%\pause
\vspace*{-.6cm}
Create a large \textbf{symbolic matrix} of squared distances:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
D_ij = ((x_i - y_j) ** 2).sum(dim=2)  # (1M, 1M) symbolic
\end{minted}

%\pause
\vspace*{-.6cm}
Use an \texttt{.argmin()} \textbf{reduction}to perform a nearest neighbor query:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
indices_i = D_ij.argmin(dim=1)  # -> standard torch tensor
\end{minted}

\end{frame}


\begin{frame}[fragile]{The KeOps library combines performance with flexibility}

\begin{center}

```

Script of the previous slide = efficient nearest neighbor query,\
**on par** with the bruteforce CUDA scheme of the **FAISS** library...\
And can be used with **any metric**!

```{=latex}
\end{center}

\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4]{python}
D_ij = ((x_i - x_j) ** 2).sum(dim=2)     # Euclidean
M_ij =  (x_i - x_j).abs().sum(dim=2)     # Manhattan
C_ij = 1 - (x_i | x_j)                   # Cosine
H_ij = D_ij / (x_i[...,0] * x_j[...,0])  # Hyperbolic
\end{minted}

\vspace*{-.4cm}
```

KeOps supports arbitrary **formulas** and **variables** with:

- **Reductions:** sum, log-sum-exp, K-min, matrix-vector product, etc.
- **Operations:** $+$, $\times$, sqrt, exp, neural networks, etc.
- **Advanced schemes:** batch processing, block sparsity, etc.
- **Automatic differentiation:** seamless integration with PyTorch.

```{=latex}
\end{frame}

\begin{frame}{KeOps lets users work with millions of points at a time}

\begin{center}
	Benchmark of a Gaussian \textbf{convolution} $~~a_i\gets \sum_{j=1}^N \exp(-\|x_i-y_j\|^2_{\mathbb{R}^3})\, b_j$ \\
	between \textbf{clouds of N 3D points} on a A100 GPU. 
	%\vspace*{-.3cm}

	\resizebox{.95\linewidth}{!}{\input{images/benchmarks/benchmark_gaussianconv_A100.tex}}
\end{center}
```

## Yet another ML compiler?

Many impressive tools out there (Numba, Triton, Halide...):

- Focus on **generality** (software + hardware).
- Increasingly easy to use via e.g. PyTorch 2.0.


KeOps fills a different niche (a bit like cuFFT, FFTW...):

- Focus on a **single major bottleneck**: geometric interactions.
- **Agnostic** with respect to Euclidean / non-Euclidean formulas.
- Fully compatible with PyTorch, NumPy, R.
- Can actually be $\BR{used by mathematicians}$.

`\begin{center}`{=latex}

KeOps is a **bridge** between geometers (with a maths background) \
and compiler experts (with a CS background).

`\end{center}`{=latex}




# How to solve the OT problem?

## Duality: central planning with $\text{N}\text{M}$ variables $\simeq$ outsourcing with $\text{N}+\text{M}$ variables

```{=latex}

	\hspace*{.5cm}
	%
	\begin{minipage}{\linewidth}
	\vspace{-.5cm}
	\hspace*{.5cm}
	\begin{align*}
		\OT(\a{\al},\b{\be})~&=~
		\min_{\g{\pi}} ~\langle \,\g{\pi}\,,\, \v{  \C}\, \rangle , ~~\text{with $\v{ \C(\a{x_i},\b{y_j}) ~=~ \tfrac{1}{p}\|\a{x_i} - \b{y_j}\|^p }$}
		& \longrightarrow~~&\text{Assignment}\\
		&\qquad\text{s.t.}~~ \g{\pi}~\geqslant~0, ~\quad
		\g{\pi} \,\textbf{\b{1}}~=~ \a{\al}, ~\quad
		\t{\g{\pi}} \textbf{\a{1}}~=~ \b{\be}
	\end{align*} 

	\hspace*{-1cm}\vspace{1.5cm}
	\begin{tabular}{m{.2\linewidth}@{\qquad\qquad}m{.45\linewidth}m{.2\linewidth}}
		\begin{tabular}{cc}
			\includegraphics[height=2.5cm]{images/intuitions/zhukov.jpg} \\
			$\sum_{i,j} \g{\pi_{i,j}} \, \v{ \C(\a{x_i},\b{y_j}) }$
		\end{tabular}
		&
		\only<1>{\includegraphics[width=\linewidth]{images/intuitions/wireframe_1.png}}%
		%\only<3>{\includegraphics[width=\linewidth]{images/intuitions/prices_none.png}}%
		%\only<4>{\includegraphics[width=\linewidth]{images/intuitions/prices_f.png}}%
		\onslide<2->{\includegraphics[width=\linewidth]{images/intuitions/prices.png}}
		& 
		\begin{tabular}{cc}
			\onslide<2->{ \includegraphics[height=2.5cm]{images/intuitions/postman.png} } \\
			\onslide<2->{ $\a{ \sum_{i} \alpha_i f_i } + \b{ \sum_j \beta_j g_j }$ }
		\end{tabular}
	\end{tabular}
	
	\vspace{-2cm}
	\onslide<2->{
	\begin{align*}
		\phantom{\OT(\a{\al},\b{\be})~}
		&\onslide<3->{\mathbf{=}}~ \max_{\a{f}\,,\, \b{g}}~ \qquad \langle \,\a{\al}\,,\, \a{f}\,\rangle 
		~+~\langle\,\b{\be}\,,\, \b{g}\, \rangle 
		& \longrightarrow~~&\text{FedEx} \\
		&\qquad\text{s.t.}~~\qquad \a{f(x_i)}\,+\,\b{g(y_j)}~\leqslant~ \v{  \C(\a{x_i}, \b{y_j}) } , \qquad
	\end{align*}
	}
	\end{minipage}

\end{frame}

\begin{frame}{Being too greedy... doesn't work!}
```

`\vspace*{.5cm}\begin{center}`{=latex}

![](images/dual_optimization/algorithm_31.png){height=100%}

`\end{center}`{=latex}


## The auction algorithm: take it easy with a slackness $~\varepsilon > 0$

`\vspace*{.5cm}\begin{center}`{=latex}

![](images/dual_optimization/algorithm_32.png){height=100%}

`\end{center}`{=latex}


## The Sinkhorn algorithm: use a softmin, get a well-defined optimum

`\vspace*{.5cm}\begin{center}`{=latex}

![](images/dual_optimization/algorithm_33.png){height=100%}

`\end{center}`{=latex}


## The symmetric Sinkhorn algorithm: stay close to the diagonal if $\text{A}\simeq \text{B}$

`\vspace*{.5cm}\begin{center}`{=latex}

![](images/dual_optimization/algorithm_sym.png){height=100%}

`\end{center}`{=latex}


## Remark 1: a streamlined algorithm

One key operation -- the soft, $\BR{weighted distance transform}$:
$$\a{\forall i \in [1, \text{N}]}, ~~ \a{f(x_i)} \gets \min_{\b{y\sim \beta}} \big[ \v{\C(\a{x_i}, \b{y})} - \b{g(y)} \big]
= -\varepsilon \log \b{\sum_{j=1}^\text{M} \beta_j} \exp \tfrac{1}{\varepsilon} \big[ \b{g_j} - \v{\C(\a{x_i}, \b{y_j})}\big]~.
$$

Similar to the chamfer distance transform, convolution with a Gaussian kernel... \
Fast implementations with **pyKeOps**:

- If $\v{\C(\a{x_i}, \b{y_j})}$ is a closed formula: **bruteforce** scales to $\text{N}, \text{M} \simeq \text{100k}$ in 10ms on a GPU.
  
- If $\a{\text{A}}$ and $\b{\text{B}}$ have a low-dimensional support: \
  use a clustering and **truncation** strategy to get a x10 speed-up.

- If $\a{\text{A}}$ and $\b{\text{B}}$ are supported on a 2D or 3D grid and $\v{\C(\a{x_i}, \b{y_j}) = \tfrac{1}{2}\|\a{x_i} - \b{y_j}\|^2}$: \
  use a **separable** distance transform to get a second x10 speed-up. \
  (N.B.: FFTs run into numerical accuracy issues.)

## Remark 2: annealing works!

The **Auction/Sinkhorn** algorithms:

- Improve the dual cost by at least $~\varepsilon$ at each (early) step.
- Reach an $\varepsilon$-optimal solution with $(\max \C) ~/~ \varepsilon~$ steps.

Simple heuristic: run the optimization with $\BR{decreasing values}$ of $\varepsilon$.

`\begin{center}`{=latex}

$\varepsilon$-scaling \
$~~~ = ~~~$ **simulated annealing** \
$~~~ = ~~~$ **multiscale** strategy \
$~~~ = ~~~$ **divide and conquer**

`\end{center}`{=latex}



## Remark 3: the curse of dimensionality



::: columns


:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/dual_optimization/peaky.pdf){height=45%}

`\end{center}`{=latex}

In **low dimension**:

- $\|x-y\|$ takes large and small values.
- The OT objective is $\BR{peaky}$ wrt. $\pi$.
- $\varepsilon$-optimal solutions are **useful**. 
- $\text{OT}(\text{discrete samples}) \simeq \text{OT}(\text{underlying distributions})$

::::

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/dual_optimization/flat.pdf){height=45%}

`\end{center}`{=latex}

In **high dimension**:

- $\|x-y\|$ gets closer to a constant.
- The OT objective is $\BR{flat}$ wrt. $\pi$.
- $\varepsilon$-optimal solutions are **random**. 
- $\text{OT}(\text{discrete samples}) \neq \text{OT}(\text{underlying distributions})$

::::

:::


## To recap 80+ years of work...

Key dates for discrete optimal transport with N points:

- `\cite{kantorovich1942translocation}`{=latex}: `\defi{Dual}`{=latex} problem of Kantorovitch.
- `\cite{kuhn1955hungarian}`{=latex}: `\defi{Hungarian}`{=latex} methods in $O(\text{N}^3)$.
- `\cite{bertsekas79auction}`{=latex}: **Auction** algorithm in $O(\text{N}^2)$.
- `\cite{kosowsky1994invisible}`{=latex}: **SoftAssign** = Sinkhorn + simulated annealing, in $O(\text{N}^2)$.
- `\cite{gold1998new,TPSRPM}`{=latex}: **Robust Point Matching** = Sinkhorn as a loss.
- `\cite{CuturiSinkhorn}`{=latex}: Start of the `\contexte{GPU era.}`{=latex}
- `\cite{merigot2011multiscale,levy2015numerical,schmitzer2019stabilized}`{=latex}: **multi-scale** solvers in $O(\text{N}\log \text{N})$. \
  $~$
- `\solution{Solution,}`{=latex}today: **Multiscale Sinkhorn algorithm, on the GPU**.


`\begin{center}`{=latex}
$\Longrightarrow$ Generalized `\solution{QuickSort}`{=latex}algorithm.
`\end{center}`{=latex}


```{=latex}
\end{frame}

\begin{frame}{Visualizing $F$, $G$ and the Brenier map $\nabla F(x_i) = -\tfrac{1}{\alpha_i}\partial_{x_i} \OT(\alpha,\beta)$}
	\vspace{.3cm}
\begin{center}
	\includegraphics[height=.8\textheight]{images/truncation/it_08.png}%
\end{center}

\vspace{-.5cm}
\hspace*{5.8cm}
OT plan in 2D.
\end{frame}


\begin{frame}{Visualizing $F$, $G$ and the Brenier map $\nabla F(x_i) = -\tfrac{1}{\alpha_i}\partial_{x_i} \OT(\alpha,\beta)$}
	\vspace{.3cm}
\begin{center}
	\includegraphics<1>[height=.8\textheight]{images/scaling/it_01.png}%
	\includegraphics<2>[height=.8\textheight]{images/scaling/it_02.png}%
	\includegraphics<3>[height=.8\textheight]{images/scaling/it_03.png}%
	\includegraphics<4>[height=.8\textheight]{images/scaling/it_04.png}%
	\includegraphics<5>[height=.8\textheight]{images/scaling/it_05.png}%
	\includegraphics<6>[height=.8\textheight]{images/scaling/it_06.png}%
	\includegraphics<7>[height=.8\textheight]{images/scaling/it_07.png}%
	\includegraphics<8>[height=.8\textheight]{images/scaling/it_08.png}%
\end{center}

\vspace{-.5cm}
\hspace*{5.2cm}
Iteration %
\only<1>{0}%
\only<2>{1}%
\only<3>{2}%
\only<4>{3}%
\only<5>{4}%
\only<6>{5}%
\only<7>{6}%
\only<8>{7}%
, blur $\sigma$ = $\sqrt{\varepsilon}$ = %
\only<1>{$2^0$}%
\only<2>{$2^{-1}$}%
\only<3>{$2^{-2}$}%
\only<4>{$2^{-3}$}%
\only<5>{$2^{-4}$}%
\only<6>{$2^{-5}$}%
\only<7>{$2^{-6}$}%
\only<8>{\texttt{.01}}%
\end{frame}



\begin{frame}{Visualizing $F$, $G$ and the Brenier map $\nabla F(x_i) = -\tfrac{1}{\alpha_i}\partial_{x_i} \OT(\alpha,\beta)$}
	\vspace{.3cm}
\begin{center}
	\includegraphics<1>[height=.8\textheight]{images/truncation/it_01.png}%
	\includegraphics<2>[height=.8\textheight]{images/truncation/it_02.png}%
	\includegraphics<3>[height=.8\textheight]{images/truncation/it_03.png}%
	\includegraphics<4>[height=.8\textheight]{images/truncation/it_04.png}%
	\includegraphics<5>[height=.8\textheight]{images/truncation/it_05.png}%
	\includegraphics<6>[height=.8\textheight]{images/truncation/it_06.png}%
	\includegraphics<7>[height=.8\textheight]{images/truncation/it_07.png}%
	\includegraphics<8>[height=.8\textheight]{images/truncation/it_08.png}%
\end{center}

\vspace{-.5cm}
\hspace*{5.2cm}
Iteration %
\only<1>{0}%
\only<2>{1}%
\only<3>{2}%
\only<4>{3}%
\only<5>{4}%
\only<6>{5}%
\only<7>{6}%
\only<8>{7}%
, blur $\sigma$ = %
\only<1>{$2^0$}%
\only<2>{$2^{-1}$}%
\only<3>{$2^{-2}$}%
\only<4>{$2^{-3}$}%
\only<5>{$2^{-4}$}%
\only<6>{$2^{-5}$}%
\only<7>{$2^{-6}$}%
\only<8>{\texttt{.01}}%


```


## Scaling up optimal transport to anatomical data

```{=latex}

\begin{center} 
	Progresses of the last decade add up to a 
	$\times \textbf{100}$ - $\times \textbf{1000}$ acceleration:
	
	Sinkhorn GPU
	$\xrightarrow{\times \text{10}}$
	+ KeOps
	$\xrightarrow{\times \text{10}}$
	+ Annealing
	$\xrightarrow{\times \text{10}}$
	+ Multi-scale
	
	
	With a precision of 1\%, on a modern gaming GPU:\\~
	\end{center}

	\vspace*{-.25cm}
	\begin{tabular}{cccc}
	\begin{minipage}{.2\linewidth}
	\begin{center}
	~\\
	\texttt{pip install geomloss}\\
	+\\
	modern GPU\\ (1 000 €)
	\end{center}
	\end{minipage}
	&
	\begin{minipage}{.1\linewidth}
	~\\
	$\Longrightarrow$
	\end{minipage}
	&
	\begin{minipage}{.3\linewidth}
	\includegraphics[height=3.5cm]{images/dragon_ball/lowres.png}
	\end{minipage}
	&
	\begin{minipage}{.3\linewidth}
	\includegraphics[height=3.5cm]{images/dragon_ball/highres.png}
	\end{minipage}
	\\
	&&
	10k points in 30-50ms
	&
	\textbf{100k points in 100-200ms}
	\end{tabular}
```

## A typical example in anatomy: lung registration "Exhale -- Inhale"

`\begin{center}`{=latex}

![](images/applications/poumons.png){height=85%}

`\end{center}\vspace*{-.3cm}`{=latex}

**Complex** deformations, high **resolution** (50k--300k points), high **accuracy** ($<$ 1mm).


## Three-steps registration


`\begin{center}`{=latex}

![](images/applications/lung_registration/full.png){height=85%}

`\end{center}\vspace*{-.3cm}`{=latex}




## Wasserstein barycenters `\cite{agueh_2011}`{=latex}

```{=latex}
\vspace*{-.5cm}
\begin{align*}
	\text{Barycenter}~~\a{\al^*}
	~=~\arg \min_{\a{\al}} 
	\sum_{i=1}^4 \,\lambda_i\,\text{Loss}(\,\a{\al}\,,\, \b{\be_i}\,)~.
\end{align*}
```

::: columns

:::: {.column width=45%}

`\begin{center}`{=latex}

![](images/wasserstein/linear_barycenters.png){height=60%}

**Euclidean** barycenters. \
$\text{Loss}(\a{\al},\b{\be}) = \|\a{\al}-\b{\be}\|^2_{L^2}$

`\end{center}`{=latex}

::::

:::: {.column width=45%}

`\begin{center}`{=latex}

![](images/wasserstein/wasserstein_barycenters.png){height=60%}

**Wasserstein** barycenters. \
$\text{Loss}(\a{\al},\b{\be}) = \OT(\a{\al},\b{\be})$

`\end{center}`{=latex}

::::

:::


## Wasserstein barycenters

From a computational perspective:

- The problem is **convex** (easy) wrt. the weights.
- The support of the barycenter lies in the **convex hull** of the input distributions.

The $\BR{curse of dimensionality}$ hits hard:

- In high dimension, identifying the support can become **NP-hard**.
- In dimensions 2 and 3, we can just use a grid and recover **super fast** algorithms.
  Computing OT distances and barycenters between $\BR{density maps}$ is a solved problem.


`\begin{center}`{=latex}

$\Longrightarrow$ We can now **easily** do manifold learning with e.g. UMAP \
in Wasserstein spaces of **2D and 3D** distributions.

`\end{center}`{=latex}

## An example: Anna Song's exploration of 3D shape textures `\cite{song2022generation}`{=latex}


`\begin{center}`{=latex}

![](images/umap.png){height=90%}

`\end{center}`{=latex}


# Particle systems

## Online this week on Arxiv thanks to:

```{=latex}
\vspace*{1cm}
\centerline{
\begin{tabular}{c@{\qquad}c}
\includegraphics[height=3cm]{images/sourires/medaillons/Maciej_2.png}
&
\includegraphics[height=3cm]{images/sourires/medaillons/Antoine_2.png}\\
\textbf{Maciej Buze}
&
\textbf{Antoine Diez} \\
Heriot-Watt University
&
Kyoto University
\end{tabular}
}
```

## Original motivation: the N-body problem `\cite{galaxy}`{=latex}

`\begin{center}`{=latex}

![](images/particles/galaxy_s.jpg){height=90%}

`\end{center}`{=latex}


## Coding a simple fluid simulation is now a matter of hours `\cite{coding_adventure}`{=latex}

`\begin{center}`{=latex}

![](images/particles/coding_adventure.jpg){height=90%}

`\end{center}`{=latex}


## The material point method: Disney's Frozen `\cite{stomakhin2013material}`{=latex}

`\begin{center}`{=latex}

![](images/particles/frozen.png){height=90%}

`\end{center}`{=latex}



## How can we enforce a volume preservation constraint? `\cite{power_particles_in_cell}`{=latex}

`\begin{center}`{=latex}

![](images/particles/clumps_and_voids.png){height=90%}

`\end{center}`{=latex}



## Use power diagrams $~$ i.e. $~$  semi-discrete optimal transport


`\begin{center}`{=latex}

![](images/particles/laguerre_cells.pdf){height=90%}

`\end{center}`{=latex}


- The $f_i$'s maximize the dual objective $~\sum_{i=1}^N v_i f_i + \int_{y\in\Omega} \min_{i=0}^N [\,c_i(y) - f_i\,] \,\text{d}y$.
- **Optimality** conditions $~\Longleftrightarrow~$ $\text{Vol}(\text{Cell}_i) = v_i$.
- To **compute the cells**, the objective and its gradient:
   - If $c_i(y) = \|y - x_i\|^2$ for all cells, use a clever **grid-free** algorithm.
   - Otherwise, just use $\BR{KeOps}$.


## Power plastics `\cite{power_plastics}`{=latex}

`\begin{center}`{=latex}

![](images/particles/power_plastics_1.png){height=90%}

`\end{center}`{=latex}

## Power plastics `\cite{power_plastics}`{=latex} -- without the eye candy

`\begin{center}\vspace*{1cm}`{=latex}

![](images/particles/power_plastics_2.png){height=90%}

`\end{center}`{=latex}



## Main numerical ingredients

These simulations alternate between:

1. **Moving the particles** according to your favorite N-body model.
   
2. Computing Laguerre `\BR{cells}`{=latex} with the **correct volumes**:
   - (Multiscale) Sinkhorn for tolerance $>$ 5\%. 
   - (Quasi-)Newton for tolerance $<$ 1\%.
  
3. **Correcting** the particle positions to enforce the volume-preservation constraint:
   - Jump to the centroid of the cell.
   - Or add a spring for smoother trajectories.

`\begin{center}`{=latex}
**Next talk:** Thomas Gallouët presents a rigorous analysis with Mérigot, Lévy, etc. \
**But first:** new applications with `\BR{custom cost functions}`{=latex} (thanks KeOps).
`\end{center}`{=latex}


## Anisotropic power diagrams let us model polycrystalline metals `\cite{apd_maciej_2024}`{=latex}

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/APD_1.png){height=4.4cm}

Ellipsoids.
`\end{center}`{=latex}

::::

:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/APD_2.png){height=4.4cm}

Pixel cells.

`\end{center}`{=latex}

::::


:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/APD_3.png){height=4.4cm}

5,000 crystals in 3D.

`\end{center}`{=latex}

::::
:::


## Fit to real EBSD scan of low-carbon steel `\cite{apd_maciej_2024}`{=latex}


`\vspace*{.5cm}`{=latex}

::: columns

:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/steel_scan.png){height=5cm}

Data from Tata steel.

`\end{center}`{=latex}


::::

:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/steel_model.png){height=5cm}

Our APD model.

`\end{center}`{=latex}

::::


:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/steel_synthetic.png){height=5cm}

New synthetic image.

`\end{center}`{=latex}

::::
:::


`\begin{center}`{=latex}
We can generate new, realistic 3D images with **prescribed properties** in seconds.
`\end{center}`{=latex}



## Change the cost function to simulate hard (blue) and soft (orange) cells `\cite{diez2024optimal}`{=latex}

`\vspace*{-.1cm}`{=latex}

::: columns

:::: {.column width=48%}

`\begin{center}`{=latex}

![](images/particles/iceshot_cube_1.png){height=6.5cm}

The **raw** 100x100x100 pixel grid...

`\end{center}`{=latex}

::::


:::: {.column width=48%}

`\begin{center}`{=latex}

![](images/particles/iceshot_cube_2.png){height=6.5cm}

with some Hollywood **makeup**.

`\end{center}`{=latex}

::::
:::



## Run-and-tumble motion `\cite{diez2024optimal}`{=latex}


`\vspace*{.3cm}`{=latex}

::: columns

:::: {.column width=49%}

`\begin{center}`{=latex}

![](images/particles/tumble_disk.png){height=6cm}

2D disk.

`\end{center}`{=latex}

::::


:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/particles/tumble_bricks.png){height=6cm}

3D cube.

`\end{center}`{=latex}

::::
:::



## Fire alarm! `\cite{diez2024optimal}`{=latex}


`\vspace*{.5cm}`{=latex}

::: columns

:::: {.column width=48%}

`\begin{center}`{=latex}

![](images/particles/evacuation_1.png){height=6cm}

Hard particles `\textcolor{red}{\textbf{burn}}`{=latex}.

`\end{center}`{=latex}

::::


:::: {.column width=48%}

`\begin{center}`{=latex}

![](images/particles/evacuation_2.png){height=6cm}

Soft particles **escape**.

`\end{center}`{=latex}

::::
:::


## Self-organizing swarms of blind, incompressible swimmers `\cite{diez2024optimal}`{=latex}


`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}

![](images/particles/swarm.png){height=90%}

`\end{center}`{=latex}



## Self-organizing swarms of blind, incompressible swimmers `\cite{diez2024optimal}`{=latex}


`\vspace*{.5cm}`{=latex}

::: columns

:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/swarm_t0.png){height=5cm}

$t = 0$

`\end{center}`{=latex}


::::

:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/swarm_t400.png){height=5cm}

$t=4$

`\end{center}`{=latex}

::::


:::: {.column width=32%}

`\begin{center}`{=latex}

![](images/particles/swarm_t3000.png){height=5cm}

$t = 30$

`\end{center}`{=latex}

::::
:::


`\begin{center}`{=latex}
**Order emerges** out of blind collisions and re-alignments.
`\end{center}`{=latex}



## Surface tension `\cite{diez2024optimal}`{=latex}

`\begin{center}`{=latex}

![](images/particles/surface_tensions.png){height=90%}

`\end{center}`{=latex}

## Surface tension `\cite{diez2024optimal}`{=latex} -- playing with the energy parameters

`\begin{center}`{=latex}

![](images/particles/surface_tensions_2.png){height=90%}

`\end{center}`{=latex}



# Conclusion

## Genuine team work

```{=latex}
\vspace*{1cm}

\centerline{
\begin{tabular}{ccccc}
\includegraphics[height=2cm]{images/sourires/medaillons/Benjamin_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Joan_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Thibault_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/FX_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Gabriel_2.png}
\\
\textcolor{blue!100!black}{Benjamin Charlier}
&
\textcolor{blue!100!black}{Joan Glaunès}
&
\textcolor{red!65!black}{Thibault Séjourné}
&
\textcolor{red!65!black}{~~~F.-X. Vialard~~~}
&
\textcolor{red!65!black}{~~~Gabriel Peyré~~~}
\\[.5cm]
\includegraphics[height=2cm]{images/sourires/medaillons/Alain_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Marc.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Zhengyang.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/olga.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/hieu.png}\\
Alain Trouvé
&
\textcolor{green!60!black}{Marc Niethammer}
&
\textcolor{green!60!black}{Shen Zhengyang}
&
\textcolor{black!60!black}{Olga Mula}
&
\textcolor{black!60!black}{Hieu Do}\\[.5cm]
\end{tabular}
}

```


## Key points

- Optimal Transport = volume preservation = `\solution{generalized sorting}`{=latex}: \
  $\longrightarrow~~$ Super-fast solvers on **simple domains**, especially 2D/3D spaces.\
  $\longrightarrow~~$ **Fundamental tool** at the intersection of geometry and statistics. \
  
- **"Video-game physics"** is great for modelling: \
  $\longrightarrow~~$ `\BR{Expressive}`{=latex}, real-time simulations that you can implement \
  $\phantom{\longrightarrow}~~~$ without being a Finite Elements guru.

- GPUs are more **versatile** than you think. \
  $\longrightarrow~~$ Ongoing work to provide **fast GPU backends** to researchers, \
  $\phantom{\longrightarrow}~~~$ going beyond what Google and Facebook are ready to pay for. \

`\begin{center}`{=latex}
**2026 target** for scientific Python: **interactive, web-based** simulations à la ShaderToy.
`\end{center}`{=latex}


## Documentation and tutorials are available online

```{=latex}
\begin{center}
$\Longrightarrow$\qquad \url{www.kernel-operations.io}\qquad $\Longleftarrow$
\vspace*{.15cm}

\includegraphics[width=.5\linewidth,frame]{images/geometric_data_analysis.png}

%\vspace*{.15cm}

\url{www.jeanfeydy.com/geometric_data_analysis.pdf}
\end{center}

```



# References


